/*
 * Vic Pucknell & Carl Unsworth STFC Daresbury Aug 2020
 * 
 * with acknowledgement 
 * J.Adamczewski-Musch, gsi, 12-May_2014
 *
 */

//   To build program for FEBEX data readout
//    gcc FEBEXexecV1.c -lc /MIDAS/lib_Linux64/libmbspex.so


#include <stdio.h>
#include <errno.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdlib.h>


#include <string.h>
#include <stdarg.h>

#include "pexor_gosip.h"
#include "libmbspex.h"

/* logical numer of pexor device*/
#define PEXDEVNO 0

#define PEX_MEM_OFF       0x100000
#define PCI_BAR0_SIZE     0x800000      // in bytes

#define MAX_SFP     4    /*  per FEBEX module crate - PEXOR module  */
#define MAX_SLAVE   16   /*  per SFP   */

#define PEX_SFP_OFF       0x40000

  
  static int nsfp = MAX_SFP;
  static int nslaves[MAX_SFP] = {0,0,0,0};
  static int bufid[MAX_SFP] = {0,0,0,0};

   static int sfp_channel = 0;
   
   static int nslave = MAX_SLAVE;
   static int slaveID = 0;

   static int l_status;
   static int tok_mode = 2;    /*  set to 2 for running with wait for data  */

static long sfp_phys_mem_base[MAX_SFP];

   static size_t len;
   static off_t off;
   static unsigned long status;
   
   static int totalsize = 0;
    
    char *pa, *pe;

    int *pdat;
    
    unsigned long num_submem = 17, submem_offset = 0, submem_bytes = 0;
    

void febex_term_connection ()
{
  
    if (FEBEX_PexorEnable == 1) {
    
       if (fd_pex != 0) {close(fd_pex); fd_pex = 0;}
    }
    
    if (FEBEX_Data_In_Enable == 1) {
      
      if (FEBEX_DataInHandle != 0) {fclose (FEBEX_DataInHandle); FEBEX_DataInHandle = 0;}
       
    }
    
    if (FEBEX_Data_Out_Enable == 1) {
      
       if (FEBEX_DataOutHandle != 0) {fclose (FEBEX_DataOutHandle); FEBEX_DataOutHandle = 0;}

    }
 
}

void febex_go_connection ()
{
     int rc;
  
    if (FEBEX_Data_Out_Enable == 1) {
      
      if (FEBEX_DataOutHandle != 0) {fclose (FEBEX_DataOutHandle); FEBEX_DataOutHandle = 0;}
      
      sprintf(fname,"%s/%s",FEBEX_Data_Base,FEBEX_DataOut_File);
      FEBEX_DataOutHandle = fopen (fname,"w");
      
      if (FEBEX_DataOutHandle <= 0) {
	   printf("ERROR>> open file %s (write) \n", fname);
        } else {
	   printf ("Out File = %s\n", fname);
      }
      
    }  
  
}

void febex_init_connection () 
{
    int rc;
    int module = 0;
    
/*    obtain option values from their NetVars  */
    
    readNetInt("FEBEX_PexorEnable", 0, 1, &FEBEX_PexorEnable);
    readNetInt("FEBEX_Data_In_Enable", 0, 1, &FEBEX_Data_In_Enable);
    readNetInt("FEBEX_Data_Out_Enable", 0, 1, &FEBEX_Data_Out_Enable);
    
    readNetInt("FEBEX_DataFormat", 0, 1, &FEBEX_DataFormat);
    readNetInt("FEBEX_Do_DataRead", 0, 1, &FEBEX_Do_DataRead);
    readNetInt("FEBEX_Do_DataFormat", 0, 1, &FEBEX_Do_DataFormat);
    
    readNetString("FEBEX_DataBase", 0, 1, FEBEX_Data_Base);
    readNetString("FEBEX_DataInFile", 0, 1, FEBEX_DataIn_File);
    readNetString("FEBEX_DataOutFile", 0, 1, FEBEX_DataOut_File);
    
  printf ("febex_init_connection  %d %d %d \n", FEBEX_PexorEnable, FEBEX_Data_In_Enable, FEBEX_Data_Out_Enable);
  printf ("FEBEX_DataFormat %d:\n", FEBEX_DataFormat);
  printf ("FEBEX_Do_DataRead %d:\n", FEBEX_Do_DataRead);
  printf ("FEBEX_Do_DataFormat %d:\n", FEBEX_Do_DataFormat);
  
  printf ("FEBEX_DataBase %s:\n", FEBEX_Data_Base);
  printf ("FEBEX_DataIn_File %s:\n", FEBEX_DataIn_File);
  printf ("FEBEX_DataOutFile %s:\n", FEBEX_DataOut_File);

    
    if (FEBEX_PexorEnable == 1) {
  
      /* open pexor device */
      
       fd_pex = mbspex_open(PEXDEVNO);
       if (fd_pex < 0) {
            printf("ERROR>> open /dev/pexor%d \n", PEXDEVNO);
 //          return;
       }
    
       if(*dacq_Trace > 1) printf ("pexor device /dev/pexor%d opened  %d \n", PEXDEVNO, fd_pex);
    
    }
    
    if (FEBEX_Data_In_Enable == 1) {
      
      sprintf(fname,"%s/%s",FEBEX_Data_Base,FEBEX_DataIn_File);
      FEBEX_DataInHandle = fopen (fname,"r");
      if (FEBEX_DataInHandle < 0) {printf("ERROR>> open file %s (read) \n", fname);} else {printf ("In File = %s\n", fname);}
      
    }
    
    if (FEBEX_Data_Out_Enable == 1) {
      
//    see action in febex_go_connection
      
//      sprintf(fname,"%s/%s",FEBEX_Data_Base,FEBEX_DataOut_File);
//      FEBEX_DataOutHandle = fopen (fname,"w");
//      if (FEBEX_DataOutHandle < 0) {printf("ERROR>> open file %s (write) \n", fname);} else {printf ("Out File = %s\n", fname);}
      
    }
    
    readNetInt("FEBEX.Link0NumBoards", 0, 1 , &nslaves[0]); 
    printf ("%d, SFP id: %d  slaves %d\n",rc, 0, nslaves[0]);
    
    readNetInt("FEBEX.Link1NumBoards", 0, 1 , &nslaves[1]); 
    printf ("%d, SFP id: %d  slaves %d\n",rc, 1, nslaves[1]);

    readNetInt("FEBEX.Link2NumBoards", 0, 1 , &nslaves[2]); 
    printf ("%d, SFP id: %d  slaves %d\n",rc, 2, nslaves[2]);

    readNetInt("FEBEX.Link3NumBoards", 0, 1 , &nslaves[3]); 
    printf ("%d, SFP id: %d  slaves %d\n",rc, 3, nslaves[3]);
    
    if (FEBEX_PexorEnable == 1) {

   for (sfp_channel=0; sfp_channel < nsfp; sfp_channel++)
   {
      if (nslaves[sfp_channel] != 0)
      {
	 sfp_phys_mem_base[sfp_channel] = (long) PEX_MEM_OFF + (long)(PEX_SFP_OFF * sfp_channel);
      
          if(*dacq_Trace > 1) printf ("SFP id: %d Pexor SFP physical memory base 0x%08x \n", sfp_channel, sfp_phys_mem_base[sfp_channel]);
      }
   }

    // map pipe memory
    //  these options for pipe memory
    len = (size_t) PCI_BAR0_SIZE;
    off = 0;
    off = (off_t) 0x40000000;
    int prot = PROT_WRITE | PROT_READ;
    int flags = MAP_SHARED | MAP_LOCKED;

    if(*dacq_Trace > 1) printf("map pipe at 0x%x, size: 0x%x, ...\n", off, len);
    if ((pa = (char *) mmap(NULL, len, prot, flags, fd_pex, off)) == MAP_FAILED) {
        if(*dacq_Trace > 1) printf("failed to mmap pipe at pex, return: 0x%x, %d \n", pa, pa);
        perror("mmap");
    } else {
        pe = pa + len - 1;
        if(*dacq_Trace > 1) printf("physical address:             0x%x \n", off);
        if(*dacq_Trace > 1) printf("size:                         0x%x \n", len);
        if(*dacq_Trace > 1) printf("first mapped virtual address: 0x%x \n", pa);
        if(*dacq_Trace > 1) printf("last  mapped virtual address: 0x%x \n", pe);
        pdat = (int *) pa;
    }
    
    l_status = mbspex_reset(fd_pex);
    if (l_status != 0) {
        if(*dacq_Trace > 1) printf("\n\nError %d in mbspex_reset . exiting...\n", l_status);
        febex_term_connection();
    }
    
    for (sfp_channel=0; sfp_channel< nsfp; sfp_channel++)
    {
      if (nslaves[sfp_channel] != 0)
      {
        l_status = mbspex_slave_init(fd_pex, sfp_channel, nslaves[sfp_channel]);
        if (l_status != 0) {
           if(*dacq_Trace > 1) printf("\n\nError %d in mbspex_slave_init for channel %x . exiting...\n", l_status, sfp_channel);
           febex_term_connection();
        }
        bufid[sfp_channel] = 0;
      }
    }
    
    }  /*  end of  if FEBEX_PexorEnable == 1   */
    
    FEBEX_FifoBufferPtr[module] = 0;    /*  reset */


}

void febex_readevent ()
{
    
    int l_status;
    long l_memsize;
    long l_toksize;

    long check_comm = 0, check_token = 0, check_sls = 0;
    int * pdata;
    
    int i, j;
    int module = 0;
    int rc;

/* do synchronous token request with direct dma*/
/*    returns output data in databuffer & databufferlen  */

   for (sfp_channel=0; sfp_channel < nsfp; sfp_channel++)
   {
      if (nslaves[sfp_channel] != 0)
      {

        if(*dacq_Trace > 1) printf("**** Requesting token from  sfp channel 0x%x, bufid %d, tok_mode %d ...\n", sfp_channel, bufid[sfp_channel], tok_mode);
	
	if (FEBEX_PexorEnable == 1) {
    
           if(*dacq_Trace > 1) printf("/dev/pexor %d; physical address: 0x%x \n", fd_pex, off);

           if(*dacq_Trace > 1) printf("first mapped virtual address: 0x%x \n", pa);

           l_status = mbspex_send_and_receive_tok(fd_pex, sfp_channel, tok_mode + bufid[sfp_channel], off, &dmasize, &check_comm, &check_token, &check_sls); /* set target to begin of pipe memory */
        if (l_status != 0) {
           if(*dacq_Trace > 1) printf("\n\nError %x in mbspex_send_and_receive_tok for channel %d .\n", l_status, sfp_channel);
           continue;
        }
	   
	    if(*dacq_Trace > 1) printf("+++++> send and receive token\n");
            if(*dacq_Trace > 1) printf("sfp_channel = %d; l_status = %d\n",sfp_channel, l_status);
            if(*dacq_Trace > 1) printf("tok_mode = %d; bufid = %d;  check_token = %d;  check slaves = %d\n",tok_mode,bufid[sfp_channel],check_token, check_sls);
	  
        }  else {   
	
          if (FEBEX_Data_In_Enable == 1 && FEBEX_DataInHandle >= 0) {
	  
	  rc = fread(&totalsize, 1, sizeof(totalsize), FEBEX_DataInHandle);
	  if (rc > 0) {
	  	  rc = fread(databuffer, 1, totalsize, FEBEX_DataInHandle);
		  if (rc > 0) {
		     databufferlen = totalsize/4;
		     break;
		  }
	  }
	  if (rc == 0) {
              if(*dacq_Trace > 1) printf("ERROR>> end of file %s\n", fname);
	  } else {
	      printf("ERROR>>  read file %s  => %d  %d\n", fname,rc,dmasize);
          }
	  
	  totalsize = 0;
	  databufferlen = 0;
	  
	  fclose(FEBEX_DataInHandle);
	  
          sprintf(fname,"%s/%s",FEBEX_Data_Base,FEBEX_DataIn_File);
          FEBEX_DataInHandle = fopen (fname,"r");
          if (FEBEX_DataInHandle < 0) {
              printf("ERROR>> open file %s (read) \n", fname);
	  } else {
              if(*dacq_Trace > 1) printf ("In File = %s\n", fname);
	  }
	  
          }
       }    /* end of  if (nslaves[sfp_channel] != 0)   */
	
	
       bufid[sfp_channel] = (bufid[sfp_channel] + 1) & 1;     /* update buffer toggle flag */

       if(*dacq_Trace > 1) printf("dma size on return:  0x%x = %d bytes \n", dmasize, dmasize);
       
       if (FEBEX_PexorEnable == 1) {      
       
       for (slaveID=0;  slaveID < nslaves[sfp_channel]; slaveID++)
       {
	 
	   if (*dacq_Trace > 1) {
               l_status = mbspex_slave_rd (fd_pex, sfp_channel, slaveID, REG_RST, &status);
               printf ("TOKEN>>  REG_RST gives 0x%x 0x%x \n", l_status, status);
	   }
    
           if(*dacq_Trace > 1) {
               l_memsize = mbspex_get_tok_memsize(fd_pex, sfp_channel);
               printf ( "token memory size %d \n",l_memsize);
           }

           if(*dacq_Trace > 1) {       
               l_toksize = mbspex_get_tok_datasize(fd_pex, sfp_channel, slaveID);
	       printf ( "token data size %d \n",l_toksize);
	   }
	
//	   if(*dacq_Trace > 1) {
//               l_status = mbspex_slave_rd(fd_pex, sfp_channel, slaveID, REG_DATA_LEN, &submem_bytes);
//               if (l_status != 0) printf("\n\nError %d in mbspex_slave_rd: REG_DATA_LEN \n", l_status);
//	       printf ( "channel data length: %d \n", submem_bytes);
//	   }	   
	  
	   if(*dacq_Trace > 1) {
               l_status = mbspex_slave_rd(fd_pex, sfp_channel, slaveID, REG_SUBMEM_NUM, &num_submem);
               if (l_status != 0) printf("\n\nError %d in mbspex_slave_rd: REG_SUBMEM_NUM \n", l_status);
	       printf ( "number of submem: %d \n", num_submem);
	   }
       }

//       if(*dacq_Trace > 1) printf("NumSlaves: %d, Submembytes: %d, numsubmem: %d\n",nslaves[sfp_channel],submem_bytes,num_submem);

//       totalsize = nslaves[sfp_channel] * submem_bytes * num_submem;

       if(*dacq_Trace > 1) printf("Received %d DMA bytes\n", dmasize);
       
       if (dmasize <= 0 || dmasize > DMASIZE) {
	   printf("ERROR>> returned data buffer size invalid %d <= 0 || %d > %d\n", dmasize, dmasize, DMASIZE);
	   totalsize = DMASIZE;
       } else {
	   totalsize = dmasize;
       }
      
       if (*dacq_Trace > 1) printf("Using total data size = %d bytes\n", totalsize);

       for (j=0;j<totalsize/4;j++) {databuffer[j]=pdat[j];}   /* nasty kludge for now!!!   NEED TO COMBINE CHANNELS */ 
       databufferlen = totalsize/4;
       
    }   /*  end of  if FEBEX_PexorEnable == 1   */
    
    if (FEBEX_Data_Out_Enable == 1 && FEBEX_DataOutHandle >= 0 && totalsize > 0) {

      rc = fwrite (&totalsize, 1, sizeof (totalsize), FEBEX_DataOutHandle);
      if (rc < 0) {
	    printf("ERROR>> write file %s \n", fname);
	    *dacq_State = STOPPED;
      }
      
      rc = fwrite (databuffer, 1, totalsize, FEBEX_DataOutHandle);
      if (rc < 0) {
	    printf("ERROR>> write file %s \n", fname);
	    *dacq_State = STOPPED;
      }
      
    }
       
//    put this data item into FIFO buffer for diagnostics

        FEBEX_FifoBufferPtr[module] = 0;    /*  reset */
        for (i=0; i<(FIFOBUFFERLENGTH-1); i++) {
 	   FEBEX_FifoBuffer[module][i] = 0;
	}
       
        FEBEX_FifoBuffer[module][FEBEX_FifoBufferPtr[module]] = 0xa5a5a5a5;
        FEBEX_FifoBufferPtr[module]++;
	
        FEBEX_FifoBuffer[module][FEBEX_FifoBufferPtr[module]] = totalsize;
	FEBEX_FifoBufferPtr[module]++;
       
        for (i=0; i<(FIFOBUFFERLENGTH-1); i++) {
 	   FEBEX_FifoBuffer[module][FEBEX_FifoBufferPtr[module]] = databuffer[i];
	   if (++FEBEX_FifoBufferPtr[module] == FIFOBUFFERLENGTH) FEBEX_FifoBufferPtr[module] = 0;
	}
        FEBEX_FifoBuffer[module][FEBEX_FifoBufferPtr[module]] = 0x5a5a5a5a;
        if (++FEBEX_FifoBufferPtr[module] == FIFOBUFFERLENGTH) FEBEX_FifoBufferPtr[module] = 0;

     }      // if (nslaves[sfp_channel] != 0)
     
   }        // for (sfp_channel=0; sfp_channel < nsfp; sfp_channel++)
   
}

void  febex_readdata() {   /*  called from FEBEX_interrupt in interrupt.c  */

    int i;
    int module = 0;
    
    struct timespec timedelay;
    timedelay.tv_sec = 0;
    timedelay.tv_nsec = 1000000;    /* 1 millisec  */
   
    evntbuffer = (unsigned int *) EventBufferAddr;   /* eventual output */
    evntbufferlen = 0;

    sigtrapset ("Error reading FEBEX modules");
    
    if (((*dacq_State) & 255) == GOING) {
        
         if (FEBEX_Do_DataRead == 1) {febex_readevent();}
	 
	 if (FEBEX_Do_DataFormat == 1 && totalsize > 0)  {
	   
	    febex_formatevent();
	 
//    event now in event data buffer
//    We handle here rather than FEBEX_interrupt to avoid "events" exceeding 64K bytes
	 
	    sigtrapset ("Error during histogramming");
	    FEBEXhistogram_data();  /* possibly histogram the input data */
	 
	    sigtrapset ("error in xfer procedures");
	    xfer_event(0);         /* possibly send for storage */
	 }
	
    } else {   /* STOPPED  */
    
      timedelay.tv_sec = 0;
      timedelay.tv_nsec = 10000000;
      nanosleep (&timedelay, NULL);    /*  10 millisec  */
      
    }
    
if(*dacq_Trace > 1)printf ("10 End of FEBEX_readdata %d\n", EventBufferLen);
    
    sigtrapset ("Error detected during data acquisition");

    return;

}




