


/*
    This code insert is the Statistics thread
*/


#ifndef WIN32
#include <pthread.h>
#endif

#ifdef LYNXOS
#include <time.h>
#include <resource.h>
#else
#ifdef WIN32
#else
#include <unistd.h>
#endif
#endif

#ifndef EXCSTATHIST
#include "DACQ_statHist.c"
#endif

#ifdef INCHITHIST
#include "DACQ_hitHist.c"
#endif


/* --------------------------------------------------------------------- */


#ifndef WIN32
void *DACQ_Stats(void * arg)
#else
DWORD WINAPI DACQ_Stats(LPVOID arg)
#endif
{
   int Rates[NUMCOUNTERS];
   int * RatesMem[NUMCOUNTERS];
   int Counts[NUMCOUNTERS];
   int i, j;
   int *delay;
   int *dacq_State;
   int maxcounters, numcounters;

#ifdef LYNXOS
   int id, prty1, prty2;
#endif

#ifdef WIN32
    char msgBuf[BUF_SIZE];
    DWORD dwChars;
#endif

#ifndef WIN32
     printf("Statistics thread starting\n");
#else
    sprintf(msgBuf,"Statistics thread starting\n");
    WriteConsole(hStdout, msgBuf, strlen(msgBuf), &dwChars, NULL);
#endif

    signal_block();       //    mask off signals handled by the main thread

//   adjust priority - run just above main data acquisition program

#ifdef LYNXOS
          id = BUILDPID(getpid(), getstid());
          prty1 = getpriority(PRIO_PROCESS, id);
          (void) setpriority(PRIO_PROCESS, id, prty1 + 2);
          prty2 = getpriority(PRIO_PROCESS, id);
          printf("Stats thread priority changed from %d to %d\n", prty1, prty2);
#else
#ifdef WIN32
    if (!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST))  ErrorHandler(TEXT("SetThreadPriority"));
#else
     nice(-2);
#endif
#endif

//    initialise Stats Rates NetVar arrays

   numcounters = NUMCOUNTERS;
   maxcounters = MAXCOUNTERS;

   if (numcounters == maxcounters) {
       for(i=0;i<numcounters;i++) {
           Counts[i]=Rates[i]=0;
           StatsMem[i] = (int *)getNetIntArrayMem("DACQ_Stats",i);
           RatesMem[i] = (int *)getNetIntArrayMem("DACQ_Rates",i);
       }
   } else {
       readNetInt("DACQ_Stats", 0, 1, &ScratchNetInt); 
       for (i=1; i<(numcounters/maxcounters); i++) {
          sprintf(message,"DACQ_Stats%d",i);
          readNetInt(message, 0, 1, &ScratchNetInt);
       }
       for(i=0;i<numcounters;i++) {
           Counts[i]=Rates[i]=0;
           if (i < maxcounters) {
               StatsMem[i] = (int *)getNetIntArrayMem("DACQ_Stats",i);
               RatesMem[i] = (int *)getNetIntArrayMem("DACQ_Rates",i);
           } else {
               j = i-((i/maxcounters)*maxcounters);
               sprintf(message,"DACQ_Stats%d",(i/maxcounters));
               StatsMem[i] = (int *)getNetIntArrayMem(message,j);
               sprintf(message,"DACQ_Rates%d",(i/maxcounters));
               RatesMem[i] = (int *)getNetIntArrayMem(message,j);
           }
       }
   }

   dacq_State = (int *)getNetIntMem("DACQ_State");   /* data acquisition state */
   delay = (int *)getNetIntMem("DACQ_StatsTime");

  for (;;)
  {

#ifdef LYNXOS
     taskDelay((*delay)*60);   /*  delay units are seconds */
#else
#ifdef WIN32
     Sleep((*delay)*1000);
#else
     usleep ((*delay)*1000000);
#endif
#endif

     if (*dacq_State != GOING) {

         for(i=0;i<NUMCOUNTERS;i++) {*RatesMem[i] = 0;};

         continue;
     }

/*     update the netvars arrays  */
    for(i=0;i<NUMCOUNTERS;i++) {*StatsMem[i] = Stats[i];}

//    update rates
     for(i=0;i<NUMCOUNTERS;i++) {Rates[i]=(Stats[i]-Counts[i])/(*delay); Counts[i]=Stats[i];}

//    and update the netvars arrays
     for(i=0;i<NUMCOUNTERS;i++) {*RatesMem[i] = Rates[i];}
     
  }
}


/* --------------------------------------------------------------------- */

void DACQ_InitStats()
{

#ifndef WIN32
   pthread_t ThreadID;
   pthread_attr_t ThreadAttr;
   int err;

/*    create a thread for the statistics creation  */

       (void) pthread_attr_init(&ThreadAttr);
       (void) pthread_attr_setdetachstate(&ThreadAttr,PTHREAD_CREATE_DETACHED);

       err = pthread_create(&ThreadID, &ThreadAttr, DACQ_Stats, NULL);
       if (err != 0) {
           perror("Error in pthread_create:");
           printf("Failed to start stats thread: This must be fixed\n");
           exit(16);
       }

      printf("Statistics thread created\n");
#else

    char msgBuf[BUF_SIZE];
    DWORD dwChars;

         hThreadArray[STATISTICS] = CreateThread(
              NULL,                         // default security attributes
              0,                            // use default stack size
              DACQ_Stats,                   // thread function name
              NULL,                         // argument to thread function
              0,                            // use default creation flags
              &dwThreadIdArray[STATISTICS]);                        // returns the thread identifier

         if (hThreadArray[STATISTICS] == NULL) {
             ErrorHandler(TEXT("CreateThread for Stats"));
             ExitProcess(1);
         }

      sprintf(msgBuf, "Statistics thread created (tid=%d)\n", dwThreadIdArray[STATISTICS]);
    WriteConsole(hStdout, msgBuf, strlen(msgBuf), &dwChars, NULL);

         ThreadInUse[STATISTICS] = 1;
         NumThreads++;

#endif

}

/* --------------------------------------------------------------------- */

void DACQ_ClearStats()
{
   int i;

/*    clear stats counters   */

   for(i=0;i<NUMCOUNTERS;i++) {*StatsMem[i] = Stats[i] = 0;}

  printf("Data Acquisition Statistics counters now cleared\n");
}

/* --------------------------------------------------------------------- */

void DACQ_UpdateStats()
{
   int i;

/*     update the netvars arrays  */

    for(i=1;i<NUMCOUNTERS;i++) {*StatsMem[i] = Stats[i];}
}

/* --------------------------------------------------------------------- */
